/*
 * Multi-TRRT.cpp
 *
 *  Created on: Nov 29, 2012
 *      Author: ddevaurs
 */

#include "Planner-pkg.h"

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"
#include "planner/planEnvironment.hpp"
#include "planner/Diffusion/Variants/Transition-RRT.hpp"

#include "planner/TrajectoryOptim/Classic/costOptimization.hpp"

#include "Multi-TRRT.hpp"

using namespace std;
using namespace API;


/**
 * Initialize the Multi T-RRT.
 * @return the number of nodes added to the graph
 */
unsigned MultiTRRT::init()
{
    // initialize the list of seeds
    seeds.push_back(_q_start);

    // add the additional configurations (defined in the p3d file) as seeds
    for (unsigned i(0); i < (unsigned) _Robot->getRobotStruct()->nconf; ++i)
    {
        // update the list of seeds
        confPtr_t conf(new Configuration(_Robot, _Robot->getRobotStruct()->conf[i]->q));
        seeds.push_back(conf);

        // update the graph
        Node * node = new Node(_Graph, conf);
        _Graph->insertNode(node);
        ++m_nbInitNodes;
    }

    // finalize the list of seeds
    seeds.push_back(_q_goal);

    // TODO DD Adaptive selection strategy
//    vector<ConnectedComponent*> components = _Graph->getConnectedComponents();
//    for (unsigned i(0); i < components.size(); ++i)
//    {
//    	components[i]->k = 0;
//    	components[i]->w = 0.;
//    	components[i]->p = 1. / components.size();
//    }

    // create the T-RRT expansion
    expansion = new TransitionExpansion(_Graph, _Goal);

    // initialize the Tree Planner
    return TreePlanner::init();
}


/**
 * Check whether one of the stopping conditions is satisfied.
 * @return TRUE if one of them is satisfied
 */
bool MultiTRRT::checkStopConditions()
{
    if (PlanEnv->getBool(PlanParam::stopPlanner)) {
        cout << "Stop: planning canceled by user." << endl;
        return true;
    }

    if (PlanEnv->getBool(PlanParam::planWithTimeLimit) &&
            m_time > PlanEnv->getDouble(PlanParam::timeLimitPlanning)) {
        cout << "Failure: the time limit has been reached (" << m_time << " s)." << endl;
        return true;
    }

    if (m_nbExpansion > (unsigned) PlanEnv->getInt(PlanParam::plannerMaxIterations)) {
        cout << "Failure: the maximum number of expansions has been reached." << endl;
        return true;
    }

    if (m_nbConscutiveFailures > (unsigned) ENV.getInt(Env::NbTry)) {
        cout << "Failure: the maximum number of consecutive failures has been reached." << endl;
        return true;
    }

    if (_Graph->getNumberOfNodes() > (unsigned) ENV.getInt(Env::maxNodeCompco)) {
        cout << "Failure: the maximum number of nodes has been reached." << endl;
        return true;
    }

    // success: the graph contains a single connected component
    if (_Graph->getNumberOfCompco() == 1)
    	return true;

    return false;
}


/**
 * Main function of the Multi T-RRT.
 * @return the number of nodes added to the graph
 */
unsigned MultiTRRT::run()
{
    double ts(0.);
    m_time = 0.;
    ChronoOn();

    // check the pre-conditions
    if (!this->preConditions()) {
        m_fail = true;
        cout << "Planning stopped: preconditions are not satisfied." << endl;
        return 0;
    }

    unsigned compCounter(0); // index of the current component considered for expansion
    unsigned nbCurCreatedNodes(0); // number of nodes created by one expansion step
    unsigned nbTotCreatedNodes(0); // total number of nodes created so far
    bool unbalancedComp;
    int distance = ENV.getInt(Env::DistConfigChoice);

    // Multi T-RRT loop: start by checking the stopping conditions
    while (!this->checkStopConditions())
    {
		// choose the component that will be expanded
		vector<ConnectedComponent *> components = _Graph->getConnectedComponents();
		compCounter = (compCounter + 1) % components.size();
		Node * fromNode = components[compCounter]->getNode(0);

		// TODO DD Adaptive selection strategy: choose the component that will be expanded
//		vector<ConnectedComponent *> components = _Graph->getConnectedComponents();
//		double prob = p3d_random(0., 1.) + 10e-12;
//		double sumProb(0.);
//		compCounter = 0;
//		while (compCounter < components.size() && prob > sumProb) {
//			sumProb += components[compCounter]->p;
//			++compCounter;
//		}
//		--compCounter;
//		Node * fromNode = components[compCounter]->getNode(0);

		// determine whether the components are balanced, when relevant
		if (ENV.getBool(Env::expandBalanced))
		{
			// compute the minimum number of nodes contained in a component
			unsigned minNbNodes = components[0]->getNumberOfNodes();
			for (unsigned i(1); i < components.size(); ++i)
				if (minNbNodes > components[i]->getNumberOfNodes())
					minNbNodes = components[i]->getNumberOfNodes();

			unbalancedComp = (components[compCounter]->getNumberOfNodes() > (minNbNodes + 10));
		}

		// no expansion is performed when the connected components are unbalanced, when relevant
		if (!ENV.getBool(Env::expandBalanced) || !unbalancedComp)
		{
			++m_nbExpansion;

			// perform a single expansion step of T-RRT
			nbCurCreatedNodes = this->expandOneStep(fromNode);

			// success, if some nodes have been created
			if (nbCurCreatedNodes > 0)
			{
				nbTotCreatedNodes += nbCurCreatedNodes;
				m_nbConscutiveFailures = 0;

				// TODO DD Adaptive selection strategy
//				components[compCounter]->k++;
//				components[compCounter]->w -= (1. + components[compCounter]->w) / components[compCounter]->k;

				// COMPONENT-JUNCTION STRATEGY:
				// attempt connections toward all other components, until one is successful
//	            for (unsigned i(0); i < components.size(); ++i)
//	            {
//	            	// ignore the expanded component
//	                if (i == compCounter)
//	                    continue;
//
//	                // look for the node whose configuration is the closest to that of the last node
//	                // within the connected component toward which the connection is attempted
//	                Node * closestNode = _Graph->nearestWeightNeighbour(components[i]->getNode(0),
//	                    _Graph->getLastNode()->getConfiguration(), false, distance);
//
//	                // attempt a connection toward the other component
//	                if (this->connectNodeToComp(_Graph->getLastNode(), closestNode))
//	                    break;
//	            }

				// COMPONENT-JUNCTION STRATEGY:
				// attempt a connection toward the closest component
				Node * closestNode(NULL);
				confPtr_t lastConf = _Graph->getLastNode()->getConfiguration();
				for (unsigned i(0); i < components.size(); ++i)
				{
					// ignore the expanded component
					if (i == compCounter)
						continue;

					// look for the node whose configuration is the closest to that of the last node
					Node * tempNode = _Graph->nearestWeightNeighbour(components[i]->getNode(0),
							lastConf, false, distance);

					// check whether this node is closer than the one found so far (if there is one)
					if (!closestNode || (lastConf->dist(*tempNode->getConfiguration(), distance)
							< lastConf->dist(*closestNode->getConfiguration(), distance)))
						closestNode = tempNode;
				}
				// attempt a connection toward the closest component
				this->connectNodeToComp(_Graph->getLastNode(), closestNode);

				// redraw the scene (when relevant)
				if (!ENV.getBool(Env::drawDisabled) && ENV.getBool(Env::drawExploration))
					_draw_func();
			}
			// failure, otherwise
			else {
				m_nbFailedExpansion++;
				m_nbConscutiveFailures++;

				// TODO DD Adaptive selection strategy
//				components[compCounter]->k++;
//				components[compCounter]->w += (1. - components[compCounter]->w) / components[compCounter]->k;
			}

			// TODO DD Adaptive selection strategy
//			components = _Graph->getConnectedComponents();
//			double weightSum(0.);
//			for (unsigned i(0); i < components.size(); ++i)
//				weightSum += exp(components[i]->w / 0.8);
//			for (unsigned i(0); i < components.size(); ++i)
//				components[i]->p = exp(components[i]->w / 0.8) / weightSum;
		}
        ChronoTimes(&m_time, &ts);
    }
    ChronoOff();

    return nbTotCreatedNodes;
}


/**
 * Extract a trajectory going through all the waypoints.
 */
Trajectory * MultiTRRT::getTrajectory()
{
	// a trajectory is returned only if the graph contains a single connected component
    if (_Graph->getNumberOfCompco() == 1)
    {
    	// simple case: only two waypoints (start and goal)
        if (seeds.size() == 2)
            return _Graph->extractAStarShortestPathsTraj(seeds[0], seeds[1]);

        // general case (multiple waypoints): concatenate consecutive sub-trajectories
        else if (seeds.size() > 2)
        {
            Trajectory * trajectory = _Graph->extractAStarShortestPathsTraj(seeds[0], seeds[1]);

//            CostOptimization * optimTraj = new CostOptimization(*trajectory);
//            optimTraj->resetCostComputed();
//            optimTraj->runShortCut(PlanEnv->getInt(PlanParam::smoothMaxIterations));

            for (unsigned i(2); i < seeds.size(); ++i)
            {
            	Trajectory * trajPortion = _Graph->extractAStarShortestPathsTraj(seeds[i - 1], seeds[i]);

//                CostOptimization * optimTrajPortion = new CostOptimization(*trajPortion);
//                optimTrajPortion->resetCostComputed();
//                optimTrajPortion->runShortCut(PlanEnv->getInt(PlanParam::smoothMaxIterations));

//                if (!optimTraj->concat(*optimTrajPortion))
//                    return NULL;

                if (!trajectory->concat(*trajPortion))
                    return NULL;
            }

//            return optimTraj;

            return trajectory;
        }

        // error: less than two waypoints are present
        else
            return NULL;
    }
    else
        return NULL;
}
