/*
 * Multi-Threshold-RRT.cpp
 *
 *  Created on: Feb 28, 2013
 *      Author: ddevaurs
 */

#include "Planner-pkg.h"

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"
#include "planner/planEnvironment.hpp"
#include "planner/Diffusion/Variants/Threshold-RRT.hpp"

#include "Multi-Threshold-RRT.hpp"

using namespace std;
using namespace API;


/**
 * Initialize the Multi RRT_obst.
 * @return the number of nodes added to the graph
 */
unsigned MultiThresholdRRT::init()
{
    // initialize the list of seeds
    seeds.push_back(_q_start);

    // add the additional configurations (defined in the p3d file) as seeds
    for (unsigned i(0); i < (unsigned) _Robot->getRobotStruct()->nconf; ++i)
    {
        // update the list of seeds
        confPtr_t conf(new Configuration(_Robot, _Robot->getRobotStruct()->conf[i]->q));
        seeds.push_back(conf);

        // update the graph
        Node * newNode = new Node(_Graph, conf);
        _Graph->insertNode(newNode);
        ++m_nbInitNodes;
    }

    // finalize the list of seeds
    seeds.push_back(_q_goal);

    // create the RRT_obst expansion
    expansion = new ThresholdExpansion(_Graph, _Goal);

    // initialize the Tree Planner
    return TreePlanner::init();
}


/**
 * Check whether one of the stopping conditions is satisfied.
 * @return TRUE if one of them is satisfied
 */
bool MultiThresholdRRT::checkStopConditions()
{
    if (PlanEnv->getBool(PlanParam::stopPlanner)) {
        cout << "Stop: planning canceled by user." << endl;
        return true;
    }

    if (PlanEnv->getBool(PlanParam::planWithTimeLimit) &&
            m_time > PlanEnv->getDouble(PlanParam::timeLimitPlanning)) {
        cout << "Failure: the time limit has been reached (" << m_time << " s)." << endl;
        return true;
    }

    if (m_nbExpansion > (unsigned) PlanEnv->getInt(PlanParam::plannerMaxIterations)) {
        cout << "Failure: the maximum number of expansions has been reached." << endl;
        return true;
    }

    if (m_nbConscutiveFailures > (unsigned) ENV.getInt(Env::NbTry)) {
        cout << "Failure: the maximum number of consecutive failures has been reached." << endl;
        return true;
    }

    if (_Graph->getNumberOfNodes() > (unsigned) ENV.getInt(Env::maxNodeCompco)) {
        cout << "Failure: the maximum number of nodes has been reached." << endl;
        return true;
    }

    // success: the graph contains a single connected component
    if (_Graph->getNumberOfCompco() == 1)
    	return true;

    return false;
}


/**
 * Main function of the Multi RRT_obst.
 * @return the number of nodes added to the graph
 */
unsigned MultiThresholdRRT::run()
{
    double ts(0.);
    m_time = 0.;
    ChronoOn();

    // check the pre-conditions
    if (!this->preConditions()) {
        m_fail = true;
        cout << "Planning stopped: preconditions are not satisfied." << endl;
        return 0;
    }

    unsigned compCounter(0); // index of the current component considered for expansion
    unsigned nbCurCreatedNodes(0);  // number of nodes created by one expansion step
    unsigned nbTotCreatedNodes(0);  // total number of nodes created so far
    int distance = ENV.getInt(Env::DistConfigChoice);

    // Multi Threshold-RRT loop: start by checking the stopping conditions
    while (!this->checkStopConditions())
    {
		++m_nbExpansion;

		// choose the component that will be expanded
		vector<ConnectedComponent *> components = _Graph->getConnectedComponents();
		compCounter = (compCounter + 1) % components.size();
		Node * fromNode = components[compCounter]->getNode(0);

		// perform a single expansion step of T-RRT
		nbCurCreatedNodes = this->expandOneStep(fromNode);

		// success, if some nodes have been created
		if (nbCurCreatedNodes > 0)
		{
			nbTotCreatedNodes += nbCurCreatedNodes;
			m_nbConscutiveFailures = 0;

			// look for the closest component
			Node * closestNode(NULL);
			confPtr_t lastConf = _Graph->getLastNode()->getConfiguration();
			for (unsigned i(0); i < components.size(); ++i)
			{
				// ignore the expanded component
				if (i == compCounter)
					continue;

				// look for the node whose configuration is the closest to that of the last node
				Node * tempNode = _Graph->nearestWeightNeighbour(components[i]->getNode(0),
						lastConf, false, distance);

				// check whether this node is closer than the one found so far (if there is one)
				if (!closestNode || (lastConf->dist(*tempNode->getConfiguration(), distance)
						< lastConf->dist(*closestNode->getConfiguration(), distance)))
					closestNode = tempNode;
			}

			// attempt a link with the closest component
			this->connectNodeToComp(_Graph->getLastNode(), closestNode);

			// redraw the scene (when relevant)
			if (!ENV.getBool(Env::drawDisabled) && ENV.getBool(Env::drawExploration))
				_draw_func();
		}
		// failure, otherwise
		else {
			m_nbFailedExpansion++;
			m_nbConscutiveFailures++;
		}

        ChronoTimes(&m_time, &ts);
    }
    ChronoOff();

    return nbTotCreatedNodes;
}


/**
 * Extract a trajectory going through all the waypoints.
 */
Trajectory * MultiThresholdRRT::extractTrajectory()
{
	// a trajectory is returned only if the graph contains a single connected component
    if (_Graph->getNumberOfCompco() == 1)
    {
    	// simple case: only two waypoints (start and goal)
        if (seeds.size() == 2)
            return _Graph->extractAStarShortestPathsTraj(seeds[0], seeds[1]);

        // general case (multiple waypoints): concatenate consecutive sub-trajectories
        else if (seeds.size() > 2)
        {
            Trajectory * trajectory = _Graph->extractAStarShortestPathsTraj(seeds[0], seeds[1]);

            for (unsigned i(2); i < seeds.size(); ++i)
                if (!trajectory->concat(*_Graph->extractAStarShortestPathsTraj(seeds[i-1], seeds[i])))
                    return NULL;

            return trajectory;
        }

        // error: less than two waypoints are present
        else
            return NULL;
    }
    else
        return NULL;
}

